from pathlib import Path
from uuid import uuid4

from markdownup.config import Config


class Entry:

    def __init__(self, context, path: Path = None, depth: int = 0):

        path = path or context.root_path

        print(f'Processing {path}')

        self.context = context
        self.config: Config = context.config
        self.path: Path = path
        self.depth = depth
        self.uid = uuid4()
        self.name = path.name
        self.request_path = None
        self.has_children = False



from abc import ABC
from typing import Optional

from markdownup.response import Response


class AuthProvider(ABC):

    @staticmethod
    def instance(context):
        auth_config = context.config.get('access', 'auth')
        if auth_config:
            auth_type = auth_config.get('type', None)
            if auth_type is None:
                raise ValueError('Config auth section found but type key is missing')
            elif auth_type == 'keycloak':
                from markdownup.auth.keycloak.keycloak import Keycloak
                return Keycloak(context)
            elif auth_type == 'basic-auth':
                from markdownup.auth.basicauth.basic_auth import BasicAuth
                return BasicAuth(context)
            else:
                raise ValueError('Unknown auth type: ' + auth_type)
        else:
            return None

    def handle_request(self, environ) -> Optional[Response]:
        return None



from pathlib import Path

import yaml

from markdownup.helpers import rget

default_config = {
    'theme': 'default',
    'wsgi': {
        'bind': '0.0.0.0:8080',
        'workers': 4
    },
    'cache': {
        'type': 'builtin',
        'bind': '127.0.0.1:8081'
    },
    'content': {
        'root': '.',
        'exclusions': [
            r'.*/\.'  # exclude hidden files and directories by default
        ],
        'indices': [
            'README.md',
            'index.md'
        ],
        'gits': {}
    },
    'access': {
        'filename': '.upaccess'
    },
    'markdown': {
        'extensions': {
            'extra': {},
            'codehilite': {
                'guess_lang': False
            }
        }
    }
}


class Config:

    @classmethod
    def from_dict(cls, dictionary):
        return cls(dictionary)

    @classmethod
    def from_file(cls, file: Path):
        return cls(yaml.load(file.read_text(), yaml.FullLoader))

    def __init__(self, custom_config):
        self.config = custom_config

    def get(self, *args):
        try:
            return rget(self.config or default_config, *args)
        except KeyError:
            try:
                return rget(default_config, *args)
            except KeyError:
                return None



import os
import re
import subprocess
from pathlib import Path

import chevron
import markdown
from markdownup.filesystem.entry import Entry
from markdownup.filesystem.file import File
from markdownup.response import Response


class MarkdownFile(Entry, File):

    _title_pattern = re.compile(r'^#\s?(.*)', re.MULTILINE)
    _search_term_pattern = re.compile(r'\w{2,}')

    def __init__(self, context, path: Path, depth: int):

        super().__init__(context, path, depth)

        self.name = self._get_title(path.read_text()) or self.name
        self.request_path = '/' + '/'.join(path.parts[len(path.parts) - depth - 1:])
        self.version = self._get_version_details()
        self.search_index = self._get_search_index()

    def get_response(self, environ):

        # relying on the fact that no process handles multiple requests at the same time, we can safely do this
        return_dir = os.getcwd()
        os.chdir(self.path.parent)
        self.context.root.apply_access(environ)

        html = chevron.render(
            template=self.context.theme.html['document'],
            partials_dict=self.context.theme.html,
            data={
                'title': self.name,
                'file': self,
                'content': markdown.markdown(
                    self.path.read_text(),
                    extensions=self.config.get('markdown', 'extensions').keys(),
                    extension_configs=self.config.get('markdown', 'extensions')
                ),
                'root': self.context.root,
                'auth': environ.get('auth', None),
                'config': self.context.config.get('render')
            }
        )

        os.chdir(return_dir)

        return Response(
            '200 OK',
            [('Content-Type', 'text/html')],
            (bytes(b, 'UTF-8') for b in html.splitlines(keepends=True))
        )

    @staticmethod
    def _get_title(md: str) -> str:
        match = MarkdownFile._title_pattern.search(md)
        return match.group(1) if match else None

    def _get_version_details(self):

        # define the cwd and arg for the git call assuming .git is not external
        cwd = self.path.parent
        arg = self.path.name

        # update cwd and arg if we match a configured external .git
        root = self.context.root_path
        relative = self.path.relative_to(root)
        for root_path, git_path in self.config.get('content', 'gits').items():
            try:
                possible_arg = relative.relative_to(root_path)
                cwd = git_path
                arg = possible_arg
                break
            except ValueError:
                pass

        result = subprocess.run(
            ['git', 'log', '-1', '--format=%an%n%ae%n%cI%n%h%n%H', '--', arg],
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL
        )
        if result.returncode == 0:
            lines = result.stdout.decode('UTF-8').splitlines()
            if lines:
                return {
                    'author': {
                        'name': lines[0],
                        'email': lines[1]
                    },
                    'date': lines[2],
                    'short_hash': lines[3],
                    'hash': lines[4]
                }
            return lines
        else:
            return None

    def _get_search_index(self):
        search_index = {}
        haystack = self.path.read_text()
        index = 0
        for needle in self._search_term_pattern.finditer(haystack):
            index += 1
            search_term = needle.group(0).lower()
            if search_term not in search_index:
                search_index[search_term] = index
        return search_index



from pathlib import Path
from typing import List, Optional, Dict, Set

from markdownup.filesystem.asset_file import AssetFile
from markdownup.filesystem.entry import Entry
from markdownup.filesystem.file import File
from markdownup.filesystem.markdown_file import MarkdownFile
from markdownup.response import ResponseException


class Directory(Entry):

    def __init__(self, context, path: Path = None, depth: int = 0):

        super().__init__(context, path, depth)

        self.index: Optional[MarkdownFile] = None
        self.directory_map: Dict[str, Directory] = {}
        self.file_map: Dict[str, MarkdownFile] = {}
        self.access_roles: Optional[Set[str]] = self._read_access_file()

        for index_filename in self.config.get('content', 'indices'):
            index_path = self.path / index_filename
            if index_path.is_file():
                self.index = MarkdownFile(context, index_path, depth)
                self.name = self.index.name
                self.request_path = self.index.request_path
                break

        for entry in self.path.iterdir():
            if self.index and self.index.path == entry:
                continue
            str_path = str('/' / entry.relative_to(context.root_path))
            exclusion = next((exclusion for exclusion in context.exclusions if exclusion.search(str_path)), None)
            if exclusion:
                print(f'Excluding {entry} due to configured exclusion pattern  {exclusion.pattern}')
                continue
            name = entry.name
            if entry.is_dir():
                self.directory_map[name] = Directory(context, entry, depth + 1)
            elif entry.is_file() and name.endswith('.md'):
                self.file_map[name] = MarkdownFile(context, entry, depth)

        # these are updated for every request
        self.children = ChrevronList()
        self.traversed = False  # keeps track of having been traversed during the most recent resolve call

    def _read_access_file(self) -> Optional[Set[str]]:
        access_file_name = self.config.get('access', 'filename')
        if access_file_name:
            access_file_path = self.path / access_file_name
            if access_file_path.is_file():
                return set(access_file_path.read_text('UTF-8').splitlines())
        return None

    def resolve(self, environ, path: Path) -> Optional[File]:
        str_path = str('/' / path)
        exclusion = next((exclusion for exclusion in self.context.exclusions if exclusion.search(str_path)), None)
        if exclusion:
            print(f'Not resolving {str_path} because it matched configured exclusion pattern {exclusion.pattern}')
            return None
        self._reset()
        return self._resolve(environ, list(path.parts))

    def _resolve(self, environ, parts: List[str]) -> Optional[File]:
        if not self.is_accessible(environ):
            return None
        self.traversed = True
        if len(parts) == 0 and self.request_path:
            raise ResponseException('302 Found', [('Location', self.request_path)])
        next_part = parts.pop(0)
        if self.index and next_part == self.index.path.name:
            return self.index
        if next_part in self.file_map:
            return self.file_map[next_part]
        elif next_part in self.directory_map:
            return self.directory_map[next_part]._resolve(environ, parts)
        else:
            asset_file_path = self.path / next_part
            if asset_file_path.is_file():
                return AssetFile(asset_file_path)
        return None

    def _reset(self):
        if self.traversed:
            for directory in self.directory_map.values():
                if directory.traversed:
                    directory._reset()
        self.traversed = False

    def apply_access(self, environ):

        def propagate(d: Directory):
            if d.is_accessible(environ):
                d.apply_access(environ)
                if d.index or d.children:
                    return d

        self.children = ChrevronList()
        self.children.extend(filter(propagate, self.directory_map.values()))
        self.children.extend(self.file_map.values())
        self.children.sort(key=lambda e: e.name)
        self.has_children = len(self.children) > 0

    def is_accessible(self, environ):
        if self.access_roles is None:
            return True  # no access rules defined, access allowed
        if 'authenticated' not in environ['auth']:
            return False  # access rules defined but user has no roles, access denied
        if self.access_roles & environ['auth']['authenticated']['roles']:  # TODO find first match is probably faster
            return True  # overlap between access roles and user roles, access allowed
        else:
            return False  # no intersect between access roles and user roles, access denied

    def get_markdown_files(self, environ):
        if self.is_accessible(environ):
            self.apply_access(environ)
            yield from Directory._get_markdown_files(self)

    @staticmethod
    def _get_markdown_files(directory: 'Directory'):
        if directory.index:
            yield directory.index
        for child in directory.children:
            if isinstance(child, MarkdownFile):
                yield child
            if isinstance(child, Directory):
                yield from Directory._get_markdown_files(child)


class ChrevronList(list):
    _CHEVRON_return_scope_when_falsy = True



from abc import ABC, abstractmethod
from typing import Optional


class Cache(ABC):

    @staticmethod
    def instance(context) -> 'Cache':
        cache_type = context.config.get('cache', 'type')

        if cache_type == 'builtin':
            from markdownup.cache.builtin.builtin_cache import BuiltinCache
            return BuiltinCache(context)
        elif cache_type == 'redis':
            from markdownup.cache.redis.redis_cache import RedisCache
            return RedisCache(context)
        else:
            raise ValueError('Unknown cache type: ' + cache_type)

    @abstractmethod
    def get(self, key: str) -> Optional[bytes]:
        raise NotImplementedError()

    @abstractmethod
    def put(self, key: str, value: bytes) -> None:
        raise NotImplementedError()

    @abstractmethod
    def delete(self, key: str) -> None:
        raise NotImplementedError()



import mimetypes
from pathlib import Path

from markdownup.filesystem.file import File
from markdownup.response import Response


class AssetFile(File):

    def __init__(self, path: Path):
        self.path = path

    def get_response(self, environ) -> Response:

        guessed_mimetype = mimetypes.guess_type(self.path.name)[0]

        def reader():
            with self.path.open('rb') as f:
                while True:
                    data = f.read(1024)
                    if not data:
                        break
                    yield data

        return Response(
            '200 OK',
            [('Content-Type', guessed_mimetype or 'application/octet-stream')],
            reader()
        )



from typing import Optional

from redis import Redis

from markdownup.cache.cache import Cache
from markdownup.markdownup import MarkdownUp


class RedisCache(Cache):

    def __init__(self, context: MarkdownUp):
        self.redis: Redis = Redis(
            host=context.config.get('cache', 'host'),
            port=context.config.get('cache', 'port'),
            db=context.config.get('cache', 'db'),
        )
        self.key_prefix = context.config.get('cache', 'prefix') or 'markdownup/'

    def get(self, key: str) -> Optional[bytes]:
        return self.redis.get(self.key_prefix + key)

    def put(self, key: str, value: bytes) -> None:
        self.redis.set(self.key_prefix + key, value)

    def delete(self, key: str) -> None:
        self.redis.delete(self.key_prefix + key)



from gunicorn.app.base import BaseApplication
from markdownup.config import Config
from markdownup.markdownup import MarkdownUp


class WsgiApplication(BaseApplication):

    def __init__(self, config: Config):
        self.config = config
        super().__init__()

    def init(self, parser, options, args):
        pass

    def load_config(self):
        for key, value in self.config.get('wsgi').items():
            self.cfg.set(key, value)

    def load(self):
        return MarkdownUp(self.config).wsgi_app



import json
from http.cookies import BaseCookie
from typing import Optional
from urllib.parse import parse_qs, urlencode
from uuid import uuid4

import jwt
import requests

from markdownup.auth.auth_provider import AuthProvider
from markdownup.helpers import rget
from markdownup.markdownup import MarkdownUp
from markdownup.response import Response


class Keycloak(AuthProvider):

    def __init__(self, context: MarkdownUp):

        self.cache = context.cache
        self.config = context.config
        self.client_id = context.config.get('access', 'auth', 'client_id')
        self.display_name_keys = context.config.get('access', 'auth', 'display_name')
        self.display_name_keys = self.display_name_keys.split('.') if self.display_name_keys else None
        self.roles_keys = context.config.get('access', 'auth', 'roles') or []
        self.roles_keys = list(map(lambda s: s.split('.'), self.roles_keys))

        base_url = \
            self.config.get('access', 'auth', 'auth_url').rstrip('/') + \
            '/realms/' + \
            self.config.get('access', 'auth', 'realm') + \
            '/protocol/openid-connect'

        self.auth_url = base_url + '/auth?' + urlencode({
            'client_id': self.client_id,
            'response_type': 'code'
        })

        self.token_url = base_url + '/token'

    def handle_request(self, environ) -> Optional[Response]:

        session_id = self._get_session_id(environ)
        cache_key = 'session/' + session_id
        query = parse_qs(environ['QUERY_STRING'])

        if 'login' in query and query['login'][0] == 'yes':
            location = self.auth_url + '&' + urlencode({'redirect_uri': self._get_redirect_url(environ)})
            return self._get_redirect_response(location, session_id)

        if 'logout' in query and query['logout'][0] == 'yes':
            self.cache.delete(cache_key)
            return self._get_redirect_response(self._get_redirect_url(environ), session_id)

        # handle incoming Keycloak redirect
        if 'code' in query:

            redirect_url = self._get_redirect_url(environ)

            token_response = requests.post(
                self.token_url,
                data={
                    'client_id': self.client_id,
                    'grant_type': 'authorization_code',
                    'code': query['code'][0],
                    'redirect_uri': redirect_url
                })
            token_response.raise_for_status()

            access_token = token_response.json()['access_token']
            access_token = jwt.decode(access_token, options={'verify_signature': False})

            cache_value = {
                'display_name': rget(access_token, *self.display_name_keys) if self.display_name_keys else None,
                'roles': list(self._get_roles(access_token)),
                'access_token': access_token
            }

            self.cache.put(cache_key, json.dumps(cache_value).encode('UTF-8'))
            return self._get_redirect_response(redirect_url, session_id)

        cache_value = self.cache.get(cache_key)
        if cache_value:  # TODO and not expired
            cache_value = json.loads(cache_value)
            cache_value['roles'] = set(cache_value['roles'])
            environ['auth']['authenticated'] = cache_value

        return None

    def _get_session_id(self, environ):
        cookies = BaseCookie(environ.get('HTTP_COOKIE', None))
        session_id = cookies['session'].value if 'session' in cookies else None
        if not session_id:
            while True:
                session_id = str(uuid4())
                if not self.cache.get('session/' + session_id):
                    break
        return session_id

    def _get_redirect_url(self, environ):
        return self.config.get('access', 'auth', 'redirect_url') + (environ['PATH_INFO'] or '/')

    @staticmethod
    def _get_redirect_response(location: str, session_id):
        return Response(
            '302 Found', [
                ('Set-Cookie', f'session={session_id}'),
                ('Location', str(location))
            ], iter([]))

    def _get_roles(self, access_token):
        roles = set()
        for roles_keys in self.roles_keys:
            value = rget(access_token, *roles_keys)
            if isinstance(value, list):
                roles.update(value)
            else:
                roles.add(value)
        return roles



from typing import Dict, List

import chevron

from markdownup.filesystem.markdown_file import MarkdownFile
from markdownup.response import Response


def search(context, environ, search_terms):
    search_scores: Dict[MarkdownFile, float] = {}

    for markdown_file in context.root.get_markdown_files(environ):
        score = 0.0
        for term in search_terms:
            term_score = markdown_file.search_index.get(term.lower(), 0.0)
            if term_score:
                score += 1.0 / term_score
        if score:
            search_scores[markdown_file] = score

    search_results: List[MarkdownFile] = \
        [entry[0] for entry in sorted(search_scores.items(), key=lambda entry: entry[1], reverse=True)]

    return search_results


def get_search_response(context, environ, request):

    search_terms = request['terms'][0]
    search_results = search(context, environ, search_terms.split(' '))

    html = chevron.render(
        template=context.theme.html['search-results'],
        partials_dict=context.theme.html,
        data={
            'title': 'Search results',
            'root': context.root,
            'auth': environ.get('auth', None),
            'search_terms': search_terms,
            'search_results': search_results
        }
    )

    return Response(
        '200 OK',
        [('Content-Type', 'text/html')],
        (bytes(b, 'UTF-8') for b in html.splitlines(keepends=True))
    )



class Response:
    def __init__(self, status: str, headers=None, body=None):
        self.status = status
        self.headers = headers or []
        self.body = body

        if not body:
            self.body = iter([bytes(status, 'UTF-8')])
            self.headers.append(('Content-Type', 'text/plain'))


class ResponseException(BaseException):
    def __init__(self, status: str, headers=None, body=None):
        self.response = Response(status, headers, body)



from pathlib import Path


class Theme:

    def __init__(self, config):
        self.path = self._resolve_path(config.get('theme'))
        self.html = {html.name[:-5]: html.read_text() for html in self.path.glob('*.html')}

    @staticmethod
    def _resolve_path(theme: str):
        path = Path(theme)
        if not path.is_dir():
            path = Path(__file__).parent / 'themes' / theme
            if not path.is_dir():
                raise ValueError(f'Theme "{theme}" not found')  # TODO improve feedback
        return path



from typing import Dict

from gunicorn.app.base import BaseApplication

from markdownup.config import Config


class CacheApplication(BaseApplication):

    def __init__(self, config: Config):
        self.config = config
        self.cache: Dict[str, bytes] = {}  # TODO swap for an LRU cache
        super().__init__()

    def init(self, parser, options, args):
        pass

    def load_config(self):
        self.cfg.set('bind', self.config.get('cache', 'bind'))
        self.cfg.set('workers', 1)

    def load(self):
        return self.wsgi_app

    def wsgi_app(self, environ, start_response):

        method = environ['REQUEST_METHOD']
        key = environ['PATH_INFO'] or '/'

        if method == 'PUT':
            value = environ['wsgi.input'].read()
            self.cache[key] = value
            start_response('200 OK', [])
            yield from iter([])
        elif method == 'GET':
            value = self.cache.get(key, None)
            if value is not None:
                start_response('200 OK', [])
                yield from iter([value])
            else:
                start_response('404 Not Found', [])
                yield from iter([])
        elif method == 'DELETE':
            self.cache.pop(key, None)
            start_response('200 OK', [])
            yield from iter([])
        else:
            start_response('405 Method Not Allowed', [])
            yield from iter([])



from typing import List


def rget(d: dict, *keys):
    return _rget(d, list(keys))


def _rget(d: dict, keys: List[str]):
    d = d[keys.pop(0)]
    if len(keys):
        return _rget(d, keys)
    else:
        return d



from base64 import b64decode
from typing import Optional
from urllib.parse import parse_qs, urlencode

from markdownup.auth.auth_provider import AuthProvider
from markdownup.markdownup import MarkdownUp
from markdownup.response import Response


class BasicAuth(AuthProvider):

    def __init__(self, context: MarkdownUp):
        self.config = context.config
        self.realm = self.config.get('access', 'auth', 'realm')
        self.users = {}

        for username, details in self.config.get('access', 'auth', 'users').items():
            self.users[username] = {
                'password': details['password'],
                'roles': set(details['roles'])
            }

    def handle_request(self, environ) -> Optional[Response]:

        authorization = environ.get('HTTP_AUTHORIZATION', None)

        if authorization:

            authorization = authorization[6:]  # drop the "Basic " prefix
            authorization = b64decode(authorization)
            authorization = authorization.decode('UTF-8')
            user, password = authorization.split(':', 1)
            if user in self.users and self.users[user]['password'] == password:
                environ['auth']['authenticated'] = {
                    'display_name': user,
                    'roles': self.users[user]['roles']
                }

        query = parse_qs(environ['QUERY_STRING'])

        login_response = self.logx('login', environ, query)
        if login_response:
            return login_response

        logout_response = self.logx('logout', environ, query)
        if logout_response:
            return logout_response

        return None

    def logx(self, x: str, environ, query):

        if x in query:

            # step one, add current path to query as redirect path and redirect to /
            # this ensures the browser associates the auth credentials with the whole site
            if 'redirect_path' not in query:
                return Response(
                    '303 See Other',
                    [('Location', '/?' + urlencode({
                        x: 'yes',
                        'redirect_path': environ['PATH_INFO'] or '/'
                    }))]
                )

            # step two, as long as the presence of the authorization header does not match
            # the state described by `x` we respond with 401 Unauthorized
            if ('authenticated' not in environ['auth']) == (x == 'login'):
                return Response(
                    '401 Unauthorized',
                    [('WWW-Authenticate', f'Basic realm="{self.realm}"')]
                )

            # step three, redirect back to the redirect_path from step one
            return Response(
                '303 See Other',
                [('Location', query['redirect_path'][0])]
            )

        return None



from pathlib import Path

import setuptools

setuptools.setup(
    name='markdownup',
    description='Markdown server',
    long_description=(Path(__file__).parent / 'README.md').read_text(),
    long_description_content_type='text/markdown',
    keywords=['markdown', 'server'],
    version='0.0.41',
    author='Binokkio',
    author_email='binokkio@b.nana.technology',
    url='https://github.com/binokkio/markdownup',
    license='LGPLv3+',
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=[
        'chevron',
        'gunicorn',
        'markdown',
        'pygments',
        'pyjwt>=2',
        'pyyaml',
        'requests'
    ]
)



from typing import Optional

import requests

from markdownup.cache.cache import Cache
from markdownup.markdownup import MarkdownUp


class BuiltinCache(Cache):

    def __init__(self, context: MarkdownUp):
        self.cache_url = 'http://' + context.config.get('cache', 'bind') + '/'

    def get(self, key: str) -> Optional[bytes]:
        url = self.cache_url + key
        response = requests.get(url)
        return None if response.status_code == 404 else response.text

    def put(self, key: str, value: bytes) -> None:
        url = self.cache_url + key
        response = requests.put(url, value)
        response.raise_for_status()

    def delete(self, key: str) -> None:
        url = self.cache_url + key
        response = requests.delete(url)
        response.raise_for_status()



import re
from os.path import normpath
from pathlib import Path
from typing import List
from urllib.parse import parse_qs

import chevron
from markdown.inlinepatterns import Pattern

from markdownup.auth.auth_provider import AuthProvider
from markdownup.cache.cache import Cache
from markdownup.config import Config
from markdownup.filesystem.asset_file import AssetFile
from markdownup.filesystem.directory import Directory
from markdownup.response import Response, ResponseException
from markdownup.search.search_provider import get_search_response
from markdownup.theme import Theme


class MarkdownUp:

    def __init__(self, config: Config):
        self.config = config
        self.cache: Cache = Cache.instance(self)
        self.exclusions: List[Pattern] = list(map(lambda e: re.compile(e), config.get('content', 'exclusions')))
        self.root_path: Path = Path(config.get('content', 'root')).resolve()
        self.root: Directory = Directory(self)
        self.theme: Theme = Theme(config)
        self.auth_provider: AuthProvider = AuthProvider.instance(self)

    def wsgi_app(self, environ, start_response):
        print(environ['REQUEST_METHOD'] + ' ' + environ['PATH_INFO'])
        response = self.get_response(environ)
        start_response(response.status, response.headers)
        yield from response.body

    def get_response(self, environ):

        request_method = environ['REQUEST_METHOD']

        if self.auth_provider:
            environ['auth'] = {'available': True}
            auth_response = self.auth_provider.handle_request(environ)
            if auth_response:
                return auth_response
        else:
            environ['auth'] = {'available': False}

        if request_method == 'GET':
            return self.get(environ['PATH_INFO'] or '/', environ)
        elif request_method == 'POST':
            body = parse_qs(environ['wsgi.input'].read().decode('UTF-8'))
            action = body.get('action', ['unknown'])[0]
            if action == 'search':
                return get_search_response(self, environ, body)
            else:
                return Response('400 Bad Request')
        else:
            return Response('405 Method Not Allowed')

    def get(self, path: str, environ=None) -> 'Response':

        environ = environ or {'auth': {'available': False}}

        # normalize path
        if not path.startswith('/'):
            return Response('400 Bad Request')
        abs_path = normpath(path)
        abs_path = Path(abs_path)
        rel_path = Path(*abs_path.parts[1:])

        # serve markdown
        try:
            file = self.root.resolve(environ, rel_path)
            if file:
                return file.get_response(environ)
        except ResponseException as e:
            return e.response

        # serve theme file
        theme_file = self.theme.path / rel_path
        if theme_file.is_file():
            return AssetFile(theme_file).get_response(environ)

        # serve theme 404
        if '404' in self.theme.html:
            self.root.apply_access(environ)
            html = chevron.render(
                template=self.theme.html['404'],
                partials_dict=self.theme.html,
                data={
                    'title': '404 Not Found',
                    'root': self.root,
                    'auth': environ.get('auth', None)
                }
            )

            return Response(
                '404 Not Found',
                [('Content-Type', 'text/html')],
                (bytes(b, 'UTF-8') for b in html.splitlines(keepends=True))
            )

        # serve plain 404
        return Response('404 Not Found')



import os
import shutil
import sys
from multiprocessing.context import Process
from pathlib import Path

import yaml
from markdownup.cache.builtin.cache_application import CacheApplication
from markdownup.config import Config, default_config
from markdownup.wsgi_application import WsgiApplication


def _main():

    if len(sys.argv) == 2:

        # initialize config based on argv
        config = None
        as_path = Path(sys.argv[1])
        if as_path.is_dir():
            config = Config.from_dict({'content': {'root': sys.argv[1]}})
        elif as_path.is_file():
            config = Config.from_file(as_path)
            os.chdir(as_path.parent)  # pretend to run from the config's parent dir
        else:
            print(f'No such file or directory: {as_path}')
            exit(3)

        # launch a new process for the built in cache server, a bit crude but suffices for now
        if config.get('cache', 'type') == 'builtin':
            Process(target=CacheApplication(config).run).start()

        # TODO if configured process and cache all markdown files

        # launch the main MarkdownUp WSGI application
        WsgiApplication(config).run()

        # if the above returns we exit normally
        exit(0)

    if len(sys.argv) > 2:
        if sys.argv[1] == '--start-config':
            config = yaml.dump(default_config)
            with open(sys.argv[2], 'w') as file:
                file.write(config)
            exit(0)
        elif sys.argv[1] == '--start-theme':
            target_dir = Path(sys.argv[2])
            if target_dir.exists():
                print('Target dir exists, not doing anything')
                exit(2)
            theme_name = 'default' if len(sys.argv) == 3 else sys.argv[3]
            theme_dir = Path(__file__).parent / 'themes' / theme_name
            if not theme_dir.exists():
                theme_names = ', '.join(theme.name for theme in theme_dir.parent.iterdir() if theme.is_dir())
                print(f'No such theme "{theme_name}", themes: {theme_names}')
                exit(2)
            shutil.copytree(theme_dir, target_dir)
            exit(0)

    print('Usage option 1: <path-to-markdown-root-directory>\n'
          'Usage option 2: <path-to-config-file>\n'
          'Usage option 3: --start-config <path-to-new-file (will be overwritten if exists)>\n'
          'Usage option 4: --start-theme <path-to-new-theme> [<seed-theme>]')
    exit(1)


if __name__ == '__main__':
    _main()
